import gc
import json
import os
import warnings

import pandas as pd
import pyspark

warnings.filterwarnings('always')
warnings.filterwarnings('ignore')

from spark_datax_schema_tools.utils.utils import clean_df
from spark_datax_schema_tools.utils.utils import extract_only_parentesis
from spark_datax_schema_tools.utils.utils import extract_split_partitions
from spark_datax_schema_tools.utils.utils import spark_schema_structure2, spark_schema_structure
from spark_datax_schema_tools.utils.utils import spark_transform_dtype
from spark_datax_schema_tools.utils.utils import transform_dtype
from spark_datax_schema_tools.utils.dataframe import show_spark_df
from spark_datax_schema_tools.utils.dataframe import show_pd_df


def read_excel(path_excel=None,
               uuaa_name=None):
    if not path_excel:
        raise Exception(f'require var path_excel: {path_excel} ')

    if not uuaa_name:
        raise Exception(f'require var uuaa_name: {uuaa_name} ')

    df_summary = pd.read_excel(f'{path_excel}',
                               usecols=["UUAA", "Table", "Object version", "Frequency", "Partitions"],
                               sheet_name='Summary', squeeze=True, engine='openpyxl')
    df_summary = df_summary.fillna(method='ffill', axis=0)
    df_summary.columns = ["uuaa", "table", "object_version", "frequency", "partitions"]

    df_uuaa = pd.read_excel(f'{path_excel}', header=1, sheet_name=f'{uuaa_name}', engine='openpyxl')

    lst = [list(df_uuaa.columns)]
    df_uuaa2 = pd.DataFrame(lst, columns=list(df_uuaa.columns))
    df_uuaa2.drop(df_uuaa2.columns[df_uuaa2.columns.str.contains('unnamed', case=False)], axis=1, inplace=True)
    df_uuaa2.columns = ['table', 'naming', 'logical_Name', 'descripcion', 'format', 'key', 'mandatory', 'calculated']

    lst = [list(df_uuaa.columns)]
    columns_header = ['table', 'naming', 'logical_Name', 'descripcion', 'format', 'key', 'mandatory', 'calculated']
    df_uuaa2 = pd.DataFrame(lst, columns=list(df_uuaa.columns))
    df_uuaa2.drop(df_uuaa2.columns[df_uuaa2.columns.str.contains('unnamed', case=False)], axis=1, inplace=True)
    df_uuaa2.columns = columns_header

    df_uuaa3 = df_uuaa.copy()
    df_uuaa3.drop(df_uuaa3.columns[df_uuaa3.columns.str.contains('unnamed', case=False)], axis=1, inplace=True)
    df_uuaa3.columns = columns_header

    df_uuaa4 = pd.concat([df_uuaa2, df_uuaa3], ignore_index=True)
    df_uuaa4["key"] = df_uuaa4["key"].apply(clean_df)
    df_uuaa4["mandatory"] = df_uuaa4["mandatory"].apply(clean_df)
    df_uuaa4["calculated"] = df_uuaa4["calculated"].apply(clean_df)

    del df_uuaa4["descripcion"]

    df = pd.merge(df_uuaa4, df_summary, how='left', on=['table'])
    df.reset_index(inplace=True)
    df['_dtype'] = df['format'].apply(transform_dtype)
    df['_partitions'] = df['partitions'].apply(extract_split_partitions)

    del df_summary, df_uuaa, df_uuaa2, df_uuaa3, df_uuaa4
    gc.collect()
    return df


def generate_schema_datax(df=None,
                          uuaa_name=None,
                          table_name=None,
                          directory_output=None):
    if not isinstance(df, pd.DataFrame):
        raise Exception('require var df: {df} ')

    if not uuaa_name:
        raise Exception(f'require var uuaa_name: {uuaa_name} ')

    if not table_name:
        raise Exception(f'require var table_name: {table_name} ')

    df = df[df["table"] == f'{table_name}']

    rs_dict = dict()
    for index, row in df.iterrows():
        table = str(row['table']).lower().strip()
        namespace = table.split("_")[1]
        naming = str(row['naming']).lower().strip()
        _partitions = eval(row['_partitions'])
        _format = row['format']
        _mandatory = True if str(row['mandatory']).upper() == "YES" else False
        _dtype = row['_dtype']
        _object_version = row["object_version"]

        if str(_format).upper() == "DATE":
            _mask = "yyyy-MM-dd"
        elif str(_format).upper() == "TIMESTAMP":
            _mask = "yyyy-MM-dd HH:mm:ss.SSSSSS"
        elif str(_format).upper() == "TIME":
            _mask = "HH:mm:ss"
        else:
            _mask = ""

        if table not in rs_dict.keys():
            rs_dict[table] = dict(fields=list())

        rs_dict[table]["namespace"] = namespace
        rs_dict[table]["name"] = table
        rs_dict[table]["doc"] = ""
        rs_dict[table]["database"] = "master"
        rs_dict[table]["storageZone"] = "Masterdata"
        rs_dict[table]["storageType"] = "HDFS-Parquet"
        rs_dict[table]["physicalPath"] = namespace
        rs_dict[table]["partitions"] = _partitions
        rs_dict[table]["securityLevel"] = ""
        rs_dict[table]["objectVersion"] = _object_version
        rs_dict[table]["type"] = "record"

        fields_dict = dict()
        fields_dict["name"] = naming
        fields_dict["type"] = eval(_dtype)
        fields_dict["logicalFormat"] = _format
        fields_dict["deleted"] = False
        fields_dict["metadata"] = False
        fields_dict["default"] = ""
        fields_dict["mask"] = _mask
        fields_dict["locale"] = "pe"
        fields_dict["mandatory"] = _mandatory
        rs_dict[table]["fields"].append(fields_dict)

    if not os.path.exists(f'{directory_output}/{table_name}'):
        os.makedirs(f'{directory_output}/{table_name}')

    with open(f'{directory_output}/{table_name}/schema_{table_name}.json', 'w') as f:
        json.dump(rs_dict[f"{table_name}"], f, indent=4)

    return rs_dict


def generate_metadata_spark(df=None,
                            table_name=None,
                            directory_output=None
                            ):
    if not isinstance(df, pd.DataFrame):
        raise Exception('require var df: {df} ')

    if not table_name:
        raise Exception(f'require var table_name: {table_name} ')

    from pyspark.sql import types

    df = df[df["table"] == f'{table_name}']

    struct_list = list()
    struct_list_string = list()
    for index, row in df.iterrows():
        naming = str(row['naming']).lower().strip()
        format = row['format']
        struct_list.append(spark_schema_structure(col=naming, format=format))
        struct_list_string.append(spark_schema_structure2(col=naming, format=format))
    p_schema = types.StructType(struct_list)
    p_schema_string = types.StructType(struct_list_string)

    # with open(f'{directory_output}/{table_name}/metadata_{table_name}.json', 'w') as f:
    #     json.dump(p_schema.jsonValue(), f, indent=4)

    return p_schema, p_schema_string


def generate_parquet_spark(spark=None,
                           df=None,
                           table_name=None,
                           spark_schema=None,
                           spark_schema_string=None,
                           directory_output=None,
                           sample_parquet=None
                           ):
    if not spark:
        raise Exception(f'require object spark: {spark} ')

    if not isinstance(df, pd.DataFrame):
        raise Exception('require var df: {df} ')

    if not table_name:
        raise Exception(f'require var table_name: {table_name} ')

    if not spark_schema:
        raise Exception(f'require var spark_schema: {spark_schema} ')

    import random
    import string
    from datetime import datetime
    from dateutil.relativedelta import relativedelta
    from faker import Faker
    from pyspark.sql import functions as func

    fake = Faker()

    df = df[df["table"] == f'{table_name}']

    output2 = list()
    for _ in range(sample_parquet):
        columns_dict = dict()

        for index, row in df.iterrows():
            naming = str(row['naming']).lower().strip()
            format = row['format']
            _spark_format = spark_transform_dtype(format=format)
            _parentheses = extract_only_parentesis(format=format)

            _fake = None
            if _spark_format in ("INTEGER",):
                _fake = fake.pyint(min_value=0, max_value=9999)
            elif _spark_format in ("TIMESTAMP",):
                d2 = datetime.now()
                d1 = d2 - relativedelta(months=6)
                _fake = fake.date_time_between(start_date=d1, end_date=d2)
            elif _spark_format in ("DECIMAL",):
                _parentheses_split = str(_parentheses).split(",")
                if len(_parentheses_split) <= 1:
                    _decimal_left_digits = int(_parentheses_split[0])
                    if _decimal_left_digits >= 23:
                        _decimal_left_digits = 23
                    _decimal_right_digits = 6

                else:
                    _decimal_left_digits = int(_parentheses_split[0])
                    _decimal_right_digits = int(_parentheses_split[1])
                    if _decimal_left_digits >= 23:
                        _decimal_left_digits = 23
                        _decimal_right_digits = 6
                _fake = fake.pydecimal(left_digits=_decimal_left_digits, right_digits=_decimal_right_digits, positive=True)
            elif _spark_format in ("TIME",):
                _fake = fake.time()
            elif _spark_format in ("DATE",):
                d2 = datetime.today()
                d1 = d2 - relativedelta(months=6)
                _fake = fake.date_between(start_date=d1, end_date=d2)
            elif _spark_format in ("STRING",):
                if naming in ("g_entific_id",):
                    _fake = fake.bothify(text='PE')
                elif naming in ("gf_frequency_type", "frequency_type"):
                    _fake = fake.bothify(text='?', letters='DM')
                else:
                    _fake = ''.join(random.choices(string.ascii_letters + string.digits, k=int(_parentheses)))
            columns_dict[naming] = _fake
        output2.append(columns_dict)
    df2 = pd.DataFrame(output2)

    df3 = spark.createDataFrame(df2, schema=spark_schema_string)

    for i in spark_schema.jsonValue()["fields"]:
        column_name = str(i["name"])
        column_type = str(i["type"])
        if column_type.startswith("decimal"):
            df3 = df3.withColumn(f"{column_name}", func.col(f"{column_name}").cast(f"{column_type}"))
        elif column_type.startswith("date"):
            df3 = df3.withColumn(f"{column_name}", func.col(f"{column_name}").cast("date"))
        elif column_type.startswith("timestamp"):
            df3 = df3.withColumn(f"{column_name}", func.col(f"{column_name}").cast("timestamp"))

    df3.coalesce(1).write.mode("overwrite").parquet(f'{directory_output}/{table_name}/dummy_{table_name}.parquet')

    return df3


def generate_components(spark=None,
                        path_excel=None,
                        uuaa_name=None,
                        table_name=None,
                        directory_output="schema_datax",
                        sample_parquet=500):
    df = read_excel(path_excel=path_excel, uuaa_name=uuaa_name)

    generate_schema_datax(df=df,
                          uuaa_name=uuaa_name,
                          table_name=table_name,
                          directory_output=directory_output)

    spark_schema, spark_schema_string = generate_metadata_spark(df=df,
                                                                table_name=table_name,
                                                                directory_output=directory_output)

    df2 = generate_parquet_spark(spark=spark,
                                 df=df,
                                 table_name=table_name,
                                 spark_schema=spark_schema,
                                 spark_schema_string=spark_schema_string,
                                 directory_output=directory_output,
                                 sample_parquet=sample_parquet)

    pyspark.sql.dataframe.DataFrame.show2 = show_spark_df
    pd.DataFrame.show2 = show_pd_df

    return df2
